import copy
import enum
from multiprocessing.dummy import Value
from os import path
from tkinter import E

import valida.data
import valida.conditions as cnds
from valida.errors import (
    DuplicateRule,
    IncompatibleRules,
    MalformedContainerItemSpec,
    MalformedDataPathSpec,
)


def get_container_value_condition(
    condition,
    datum_condition,
    cls,
    cls_like,
    condition_label="condition",
    datum_condition_label="value",
):

    if condition is not None:
        if not isinstance(condition, cnds.ConditionLike):
            raise TypeError(
                f"If specified, `{condition_label}` must be a cnds.ConditionLike object."
            )
    else:
        condition = cnds.NullCondition()

    if datum_condition is not None:
        if not isinstance(datum_condition, cnds.NullCondition):
            if not isinstance(datum_condition, cnds.ConditionLike):
                datum_condition = cls.equal_to(datum_condition)
            if not datum_condition.is_like(cls_like):
                raise TypeError(
                    f"{datum_condition_label} must be a {cls.__name__} object."
                )
        condition = condition & datum_condition

    return condition


class Container(enum.Enum):

    CONTAINER = -1
    LIST = 0
    MAP = 1


class DataPathDatumType(enum.Enum):

    NONE = None
    DTYPE = 1
    LENGTH = 2
    MAP_KEYS = 3
    MAP_VALUES = 4


class DataPathMultiType(enum.Enum):

    NONE = None
    FIRST = 1  # Use just the first returned match
    LAST = 2  # Use just the final returned match
    SINGLE = 3  # As in FIRST, but raise/fail if more than one exists
    ALL = 4  #
    ANY = 5  #


class DataPath:
    """Class to represent a path within a nested data structure.

    A DataPath locates nodes within a nested data structure. It comprises an
    address of nested data types (mapping values or list items) that can be further
    refined using one ore more conditions. Validation rules are applied (i.e. tested) at
    the nodes identified by a ContainerPath.

    """

    DATUM_TYPE = DataPathDatumType.NONE
    MULTI_TYPE = DataPathMultiType.NONE

    def __init__(self, *parts, datum_type=None, multi_type=None, source_data=None):

        part_objs = []
        is_concrete = True
        for i in parts:
            if isinstance(i, ContainerValue):
                is_concrete = False
            else:
                if isinstance(i, (str, float)):
                    i = MapValue(i)
                elif isinstance(i, int):
                    i = MapOrListValue(key=i, index=i)  # could be either map or list
                else:
                    raise TypeError(
                        f"Cannot construct a DataPath from part of type: {type(i)}"
                    )
            part_objs.append(i)

        self.parts = tuple(part_objs)
        self.is_concrete = is_concrete
        self.source_data = source_data
        self.DATUM_TYPE = DataPathDatumType(datum_type)
        self.MULTI_TYPE = DataPathMultiType(multi_type)

    @property
    def DATUM_TYPE(self):
        return self._DATUM_TYPE

    @DATUM_TYPE.setter
    def DATUM_TYPE(self, value):
        self._DATUM_TYPE = DataPathDatumType(value)

    @property
    def MULTI_TYPE(self):
        return self._MULTI_TYPE

    @MULTI_TYPE.setter
    def MULTI_TYPE(self, value):
        multi_type = DataPathMultiType(value)
        if self.is_concrete and multi_type.value:
            raise ValueError(
                f"{self!r} has `MULTI_TYPE` set to {self.MULTI_TYPE.name!r}, but is a "
                f'"concrete" path for which there will only ever be at most a single '
                f"match. For concrete paths, `MULTI_TYPE` should not be set."
            )
        else:
            self._MULTI_TYPE = multi_type

    def __eq__(self, other):
        if (
            type(self) == type(other)
            and self.parts == other.parts
            and self.is_concrete == other.is_concrete
            and self.DATUM_TYPE == other.DATUM_TYPE
            and self.MULTI_TYPE == other.MULTI_TYPE
            and self.source_data == other.source_data
        ):
            return True
        return False

    @classmethod
    def from_spec(cls, spec):

        general_msg = (
            f'The specification key must start with "path" and optionally include, as '
            f'additional dot-delimited tokens, a `DATUM_TYPE` specifier (e.g. "path.map_keys", '
            f'"path.map_values", "path.type", "path.length") and/or a `MULTI_TYPE`'
            f' specifier (e.g. "path.single", "path.first", "path.last", '
            f'"path.all" or "path.single.map_keys" etc).'
        )

        DATUM_TYPE_MULTI_TYPE_LOOKUP = {
            "type": "dtype",
            "len": "length",
        }

        if not isinstance(spec, dict):
            raise MalformedDataPathSpec(general_msg)
        else:
            spec_key, spec_val = next(iter(spec.items()))  # single-item dict

        REPLACE = "path"
        ESC_CODE = rf"\{REPLACE}"
        is_escaped = False
        for k in list(spec.keys()):
            if ESC_CODE in k:
                is_escaped = True
                spec_val = spec.pop(k)
                k_new = k.replace(ESC_CODE, REPLACE)
                spec[k_new] = spec_val
        if is_escaped:
            return spec

        if len(spec) > 1:
            raise MalformedDataPathSpec(
                f"A data path should be specified with exactly one "
                f"specification key (but found keys: {list(spec.keys())}). {general_msg}"
            )

        if not isinstance(spec_key, str):
            raise MalformedDataPathSpec(general_msg)

        spec_key_split = [i.lower() for i in spec_key.split(".")]
        spec_key_split_len = len(spec_key_split)

        if spec_key_split[0] != "path" or spec_key_split_len not in range(1, 4):
            raise MalformedDataPathSpec(general_msg)

        obj = cls.from_part_specs(*spec_val)

        for i in spec_key_split[1:]:
            i = DATUM_TYPE_MULTI_TYPE_LOOKUP.get(i, i)
            try:
                obj = getattr(obj, i)()
            except AttributeError:
                raise MalformedDataPathSpec(
                    f"{i} if not a known DataPath DATUM_TYPE or MULTI_TYPE. {general_msg}"
                )

        return obj

    @classmethod
    def from_json_like(cls, json_like, *args, **kwargs):
        return cls.from_spec(json_like)

    def to_json_like(self, *args, **kwargs):
        out = self.to_part_specs()
        if "shared_data" in kwargs:
            return out, kwargs["shared_data"]
        else:
            return out

    @classmethod
    def from_part_specs(cls, *parts):
        """Construct a DataPath from parts that might include dicts, where each dict
        describes the required ContainerItem using string keys."""

        spec_resolved_parts = []
        for i in parts:
            if isinstance(i, dict):
                i = ContainerValue.from_spec(i)
            spec_resolved_parts.append(i)

        return cls(*spec_resolved_parts)

    def to_part_specs(self):
        parts = []
        for i in self.parts:
            try:
                part_spec = i.condition.callable.kwargs["value"]
            except KeyError:
                if isinstance(i, MapOrListValue):
                    part_spec = i.list_condition.callable.kwargs["value"]
            parts.append(part_spec)
        return parts

    @classmethod
    def from_str(cls, path_str, delimiter="/"):

        if path_str:
            path_str_parts = path_str.split(delimiter)
        else:
            path_str_parts = []

        path_parts = []
        for i in path_str_parts:
            try:
                i_int = int(i)  # list index (int) or map key (int or str)
                i = MapOrListValue(key=cnds.Key.in_((i, i_int)), index=i_int)
            except ValueError:
                try:
                    i_float = float(i)  # map key (float or str)
                    i = MapValue(key=cnds.Key.in_((i, i_float)))
                except ValueError:
                    pass
            path_parts.append(i)

        return cls(*path_parts)

    def __truediv__(self, other):
        """Concatenating with other MapValue, ListItem, ContainerValue or ContainerPaths."""
        if isinstance(other, ContainerValue) or other in (  # TODO: this looks wonky
            MapValue,
            ListValue,
            MapOrListValue,
        ):
            return DataPath(*self.parts, other)
        elif isinstance(other, DataPath):
            return DataPath(*self.parts, *other.parts)

    def __rtruediv__(self, other):
        return self.__class__(other) / self

    def __iter__(self):
        for i in self.parts:
            yield i

    def __len__(self):
        return len(self.parts)

    def __getitem__(self, index):
        if isinstance(index, int):
            return self.parts[index]
        elif isinstance(index, slice):
            return self.__class__(*self.parts[index])

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f'{", ".join(f"{i!r}" for i in self.parts)}'
            f'{", " if self.DATUM_TYPE.value and self.parts else ""}'
            f'{f"datum_type={self.DATUM_TYPE}" if self.DATUM_TYPE.value else ""}'
            f'{", " if self.MULTI_TYPE.value and self.parts else ""}'
            f'{f"multi_type={self.MULTI_TYPE}" if self.MULTI_TYPE.value else ""}'
            f")"
        )

    def _copy_with_datum_type(self, datum_type):
        if self.DATUM_TYPE.value:
            raise ValueError(f"{self.__class__.__name__} DATUM_TYPE is already set.")
        obj = copy.copy(self)
        obj.DATUM_TYPE = datum_type
        return obj

    def _copy_with_multi_type(self, multi_type):
        if self.MULTI_TYPE.value:
            raise ValueError(f"{self.__class__.__name__} MULTI_TYPE is already set.")
        obj = copy.copy(self)
        obj.MULTI_TYPE = multi_type
        return obj

    def dtype(self):
        return self._copy_with_datum_type(DataPathDatumType.DTYPE)

    def length(self):
        return self._copy_with_datum_type(DataPathDatumType.LENGTH)

    def map_keys(self):
        return self._copy_with_datum_type(DataPathDatumType.MAP_KEYS)

    def map_values(self):
        return self._copy_with_datum_type(DataPathDatumType.MAP_VALUES)

    def first(self):
        return self._copy_with_multi_type(DataPathMultiType.FIRST)

    def last(self):
        return self._copy_with_multi_type(DataPathMultiType.LAST)

    def single(self):
        return self._copy_with_multi_type(DataPathMultiType.SINGLE)

    def any(self):
        # TODO: how to implement?
        return self._copy_with_multi_type(DataPathMultiType.ANY)

    def all(self):
        return self._copy_with_multi_type(DataPathMultiType.ALL)

    def resolve_implicit_types(self):
        # Note: we don't support "complex mappings" from YAML where keys are themselves
        # mappings or lists
        types = []
        for i in self.parts:
            try:
                type_i = i.CONTAINER_TYPE
            except AttributeError:
                raise TypeError(f'Unknown container item type: "{type(i)}"')
            types.append(type_i)

        return types

    def _extract_specified_datum_type(self, data):
        if self.DATUM_TYPE.value:
            if self.DATUM_TYPE == DataPathDatumType.DTYPE:
                data = [type(i) for i in data]
            elif self.DATUM_TYPE == DataPathDatumType.LENGTH:
                data = [len(i) for i in data]
            elif self.DATUM_TYPE == DataPathDatumType.MAP_KEYS:
                data = [list(i.keys()) for i in data]
            elif self.DATUM_TYPE == DataPathDatumType.MAP_VALUES:
                data = [list(i.values()) for i in data]
        return data

    def _match_specified_multi_type(self, data, concrete_paths):

        if self.MULTI_TYPE.value:
            if self.MULTI_TYPE == DataPathMultiType.FIRST:
                data = data[0]
            elif self.MULTI_TYPE == DataPathMultiType.LAST:
                data = data[-1]
            elif self.MULTI_TYPE == DataPathMultiType.SINGLE:
                if len(data) > 1:
                    paths_fmt = "\n".join(
                        f"Path {idx}: {tuple(i)}"
                        for idx, i in enumerate(concrete_paths, 1)
                    )
                    raise ValueError(
                        f"{self!r} has `MULTI_TYPE` set to SINGLE, meaning there "
                        f"should be only one match, but multiple ({len(data)}) matches "
                        f"found at paths:\n{paths_fmt}."
                    )
                data = data[0]
            elif self.MULTI_TYPE == DataPathMultiType.ALL:
                pass
            # TODO: how to implement DataPathMultiType.ANY?

        elif self.is_concrete:
            # `MULTI_TYPE` will be NONE if `is_concrete`
            data = data[0]

        return data

    def get_data(self, data=None, return_paths=False):
        """Get the data specified by this path."""

        if self.source_data:
            data = self.source_data
        elif not data:
            raise ValueError("Specify the `data` from which to get sub-data.")

        if not self.parts:
            if isinstance(data, valida.data.Data):
                data = data.original
            data = self._extract_specified_datum_type([data])[0]
            if return_paths:
                return (data, ())  # if no parts, also concrete so just return one item
            else:
                return data

        data = [data]
        concrete_paths = []

        for part_idx, part in enumerate(self.parts):

            new_data = []
            new_concrete_paths = []
            for datum_idx, datum in enumerate(data):

                try:
                    filtered_data = part.filter(datum)
                except TypeError:
                    continue

                new_data.extend(filtered_data.data)

                if part_idx == 0:
                    new_concrete_paths = [[i] for i in filtered_data.keys]
                else:
                    new_concrete_paths += [
                        concrete_paths[datum_idx] + [i] for i in filtered_data.keys
                    ]

            concrete_paths = new_concrete_paths
            data = new_data

        if not data:
            return None if self.is_concrete else []

        # TODO: handle and re-raise if DATUM_TYPE processing ?
        # maybe this should raise, but it should be caught by Condition.filter and
        # fail the rule without raising an exception?
        data = self._extract_specified_datum_type(data)

        if return_paths:
            out = [(i, tuple(j)) for i, j in zip(data, concrete_paths)]
        else:
            out = data

        out = self._match_specified_multi_type(out, concrete_paths)

        return out


class ContainerValue:
    """Class for representing a container value (i.e. an item within either a list or a
    mapping). A container item can be filtered according to both its relationship to its
    parent container (i.e. position within the list for a list item, or key for a mapping
    item) and its value.

    """

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(condition={self.condition!r}"
            f'{f", label={self.label!r}" if self.label else ""}'
            f")"
        )

    def __eq__(self, other):
        if (
            type(self) == type(other)
            and self.condition == other.condition
            and self.label == other.label
        ):
            return True
        return False

    @staticmethod
    def from_spec(spec):
        CLS_LOOKUP = {
            "map_value": MapValue,
            "list_value": ListValue,
            "map_or_list_value": MapOrListValue,
        }
        container_type = spec.pop("type", "map_or_list_value")
        try:
            cls = CLS_LOOKUP[container_type]
        except KeyError:
            raise TypeError(
                f'Container item type "{container_type}" not understood. '
                f"Allowed container items are: {list(CLS_LOOKUP.keys())!r}."
            )

        # for all types:
        condition = spec.pop("condition", None)
        if condition is not None:
            condition = cnds.ConditionLike.from_spec(condition)
        else:
            condition = cnds.NullCondition()

        # for a ContainerValue type:
        list_condition = spec.pop("list_condition", None)
        if list_condition is not None:
            list_condition = cnds.ConditionLike.from_spec(list_condition)
        else:
            list_condition = cnds.NullCondition()

        # for a ContainerValue type:
        map_condition = spec.pop("map_condition", None)
        if map_condition is not None:
            map_condition = cnds.ConditionLike.from_spec(map_condition)
        else:
            map_condition = cnds.NullCondition()

        value = spec.pop("value", None)
        if value is not None:
            new_cond = cnds.ConditionLike.from_spec(value)
            if not new_cond.is_value_like:
                raise ValueError(
                    'Conditions specified in the "value" specification '
                    "must be value-like."
                )
            condition = condition & new_cond

        # shorthand specs:
        value_short_keys = [i for i in spec if i.startswith("value.")]
        value_short_cond_specs = {i: spec.pop(i) for i in value_short_keys}
        for spec_k, spec_v in value_short_cond_specs.items():
            condition = condition & cnds.ConditionLike.from_spec({spec_k: spec_v})

        if cls == MapValue:

            # shorthand specs:
            key_short_keys = [i for i in spec if i.startswith("key.")]
            key_short_cond_specs = {i: spec.pop(i) for i in key_short_keys}
            for spec_k, spec_v in key_short_cond_specs.items():
                condition = condition & cnds.ConditionLike.from_spec({spec_k: spec_v})

            key = spec.pop("key", None)
            if key is not None:
                new_cond = cnds.ConditionLike.from_spec(key)
                if not new_cond.is_key_like:
                    raise ValueError(
                        'Conditions specified in the "key" specification '
                        "must be key-like."
                    )
                condition = condition & new_cond

        elif cls == ListValue:

            # shorthand specs:
            index_short_keys = [i for i in spec if i.startswith("index.")]
            index_short_cond_specs = {i: spec.pop(i) for i in index_short_keys}
            for spec_k, spec_v in index_short_cond_specs.items():
                condition = condition & cnds.ConditionLike.from_spec({spec_k: spec_v})

            index = spec.pop("index", None)
            if index is not None:
                new_cond = cnds.ConditionLike.from_spec(index)
                if not new_cond.is_index_like:
                    raise ValueError(
                        'Conditions specified in the "index" specification '
                        "must be index-like."
                    )
                condition = condition & new_cond

        elif cls == MapOrListValue:

            # shorthand specs:
            index_short_keys = [i for i in spec if i.startswith("index.")]
            index_short_cond_specs = {i: spec.pop(i) for i in index_short_keys}
            for spec_k, spec_v in index_short_cond_specs.items():
                list_condition = list_condition & cnds.ConditionLike.from_spec(
                    {spec_k: spec_v}
                )

            # shorthand specs:
            key_short_keys = [i for i in spec if i.startswith("key.")]
            key_short_cond_specs = {i: spec.pop(i) for i in key_short_keys}
            for spec_k, spec_v in key_short_cond_specs.items():
                map_condition = map_condition & cnds.ConditionLike.from_spec(
                    {spec_k: spec_v}
                )

            index = spec.pop("index", None)
            if index is not None:
                new_cond = cnds.ConditionLike.from_spec(index)
                if not new_cond.is_index_like:
                    raise ValueError(
                        'Conditions specified in the "index" specification '
                        "must be index-like."
                    )
                list_condition = list_condition & new_cond

            key = spec.pop("key", None)
            if key is not None:
                new_cond = cnds.ConditionLike.from_spec(key)
                if not new_cond.is_key_like:
                    raise ValueError(
                        'Conditions specified in the "key" specification '
                        "must be key-like."
                    )
                map_condition = map_condition & new_cond

        label = spec.pop("label", None)
        if spec:
            raise ValueError(
                f"Unknown arguments to container item specification: {list(spec.keys())}"
            )

        if cls == MapOrListValue:
            return cls(
                condition=condition,
                list_condition=list_condition,
                map_condition=map_condition,
                label=label,
            )
        else:
            return cls(condition=condition, label=label)

    def __truediv__(self, other):
        """Concatenating with other DictValue, ListValue or DataPath objects."""
        return DataPath(self, other)

    def __rtruediv__(self, other):
        return DataPath(other) / self


class MapValue(ContainerValue):

    CONTAINER_TYPE = Container.MAP

    def __init__(self, key=None, value=None, condition=None, label=None):

        condition = get_container_value_condition(
            condition,
            key,
            cnds.Key,
            cnds.KeyLike,
            datum_condition_label="key",
        )
        condition = get_container_value_condition(
            condition,
            value,
            cnds.Value,
            cnds.ValueLike,
        )
        self.condition = condition
        self.label = label

    def filter(self, data):
        if not isinstance(data, valida.data.Data):
            data = valida.data.Data(data)
        if data.is_list:
            raise TypeError(
                "`MapValue` container can only filter a mapping (i.e. a dict)."
            )
        return self.condition.filter(data)


class ListValue(ContainerValue):

    CONTAINER_TYPE = Container.LIST

    def __init__(self, index=None, value=None, condition=None, label=None):

        condition = get_container_value_condition(
            condition,
            index,
            cnds.Index,
            cnds.IndexLike,
            datum_condition_label="index",
        )
        condition = get_container_value_condition(
            condition,
            value,
            cnds.Value,
            cnds.ValueLike,
        )
        self.condition = condition
        self.label = label

    def filter(self, data):
        if not isinstance(data, valida.data.Data):
            data = valida.data.Data(data)
        if not data.is_list:
            raise TypeError("`ListValue` container can only filter a list.")
        return self.condition.filter(data)


class MapOrListValue(ContainerValue):
    """To represent a value within a Map or a List."""

    CONTAINER_TYPE = Container.CONTAINER

    def __init__(
        self,
        key=None,
        index=None,
        value=None,
        list_condition=None,
        map_condition=None,
        condition=None,
        label=None,
    ):

        list_condition = get_container_value_condition(
            list_condition,
            index,
            cnds.Index,
            cnds.IndexLike,
            "lst_condition",
            "index",
        )
        map_condition = get_container_value_condition(
            map_condition,
            key,
            cnds.Key,
            cnds.KeyLike,
            "map_condition",
            "key",
        )
        condition = get_container_value_condition(
            condition,
            value,
            cnds.Value,
            cnds.ValueLike,
        )

        self.condition = condition
        self.list_condition = list_condition
        self.map_condition = map_condition
        self.label = label

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(condition={self.condition!r}"
            f" list_condition={self.list_condition!r}, map_condition={self.map_condition!r}"
            f'{f", label={self.label!r}" if self.label else ""}'
            f")"
        )

    def filter(self, data):
        if not isinstance(data, valida.data.Data):
            data = valida.data.Data(data)
        if data.is_list:
            condition = self.list_condition & self.condition
        else:
            condition = self.map_condition & self.condition

        return condition.filter(data)


def resolve_implicit_types(path):

    types = []
    for i in path:

        # Note: we don't support "complex mappings" from YAML where keys are themselves
        # mappings or lists
        try:
            type_i = i.CONTAINER_TYPE
        except AttributeError:
            raise TypeError(f'Unknown container item type: "{type(i)}"')

        types.append(type_i)

    return types


def validate_rule_paths(rules):

    seen_paths = []
    predicted_types = {}
    for r_idx, r in enumerate(rules):
        r_path = r["path"]
        if r_path in seen_paths:
            msg = f"Rule index {r_idx} shares with another rule the path: {r_path}"
            raise DuplicateRule(msg)
        else:
            seen_paths.append(r_path)
            # print(f'r_path: {r_path!r} is not in seen_paths:\n{seen_paths!r}\n')
        r_types = resolve_implicit_types(r_path)

        # print(f'predicted_types: {predicted_types}')

        for path_end_idx in range(len(r_path)):
            partial_path = r_path[0:path_end_idx]
            partial_path_str = f"{partial_path!r}"
            # print(path_end_idx, partial_path)
            if partial_path_str in predicted_types:
                predicted_types[partial_path_str]["rules"].append(r_idx)
                predicted_types[partial_path_str]["types"].append(r_types[path_end_idx])
            else:
                predicted_types[partial_path_str] = {
                    "rules": [r_idx],
                    "types": [r_types[path_end_idx]],
                }

    # Identify type, where possible, for each node
    for path, types_info in predicted_types.items():

        uniq_types = set(types_info["types"])

        if len(uniq_types) == 1:
            actual_type = list(uniq_types)[0]

        elif Container.LIST in uniq_types and Container.MAP in uniq_types:
            path_rules = [(i, rules[i]["path"]) for i in types_info["rules"]]
            # print(path_rules)
            path_rules_fmt = "\n\t" + "\n\t".join(
                f"Rule index {i[0]}: {i[1]!r}" for i in path_rules
            )
            msg = (
                f"Incompatible rules specified for path {path}; at least one rule "
                f"implies this node is a mapping, but at least one other rule implies "
                f"this node is a list. Associated rule paths are: {path_rules_fmt}"
            )
            raise IncompatibleRules(msg)

        elif Container.CONTAINER in uniq_types:

            if Container.LIST in uniq_types:
                actual_type = Container.LIST

            elif Container.MAP in uniq_types:
                actual_type = Container.MAP

        else:
            raise RuntimeError("Unique container types not understood.")

        predicted_types[path] = actual_type

    return predicted_types



from inspect import signature


def null_condition_binary_check(cond_1, cond_2):
    """Get the non-null condition of two conditions if one is null, otherwise return
    None."""
    return cond_1 if cond_2.is_null else (cond_2 if cond_1.is_null else None)


def get_func_args_by_kind(func):
    parameters_by_kind = {
        "POSITIONAL_OR_KEYWORD": [],
        "VAR_POSITIONAL": [],
        "VAR_KEYWORD": [],
    }
    for k, v in signature(func).parameters.items():
        try:
            parameters_by_kind[v.kind.name].append(k)
        except KeyError:
            raise NotImplementedError(
                f'Function argument with kind: "{v.kind.name}" is not allowed.'
            )

    return parameters_by_kind


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)



def cast_string_to_bool(s):
    if s.lower() == "true":
        return True
    elif s.lower() == "false":
        return False
    else:
        raise TypeError(f"Cannot cast {s!r} to a bool type.")


CAST_DTYPE_LOOKUP = {
    "str": str,
    "bool": bool,
    "int": int,
}
CAST_LOOKUP = {
    (str, bool): cast_string_to_bool,
    (str, int): int,
}



def equal_to(trial_datum, value) -> bool:
    """Returns True if the trial_datum is equal to a given value."""
    return trial_datum == value


def not_equal_to(trial_datum, value) -> bool:
    return trial_datum != value


def less_than(trial_datum, value) -> bool:
    return trial_datum < value


def greater_than(trial_datum, value) -> bool:
    return trial_datum > value


def less_than_or_equal_to(trial_datum, value) -> bool:
    return trial_datum <= value


def greater_than_or_equal_to(trial_datum, value) -> bool:
    return trial_datum >= value


def in_(trial_datum, value) -> bool:
    return trial_datum in value


def not_in(trial_datum, value) -> bool:
    return trial_datum not in value


def in_range(trial_datum, lower, upper) -> bool:
    return trial_datum in range(lower, upper)


def not_in_range(trial_datum, lower, upper):
    return trial_datum not in range(lower, upper)


def factor_of(trial_datum, value) -> bool:
    return value % trial_datum == 0


def has_factor(trial_datum, value) -> bool:
    return trial_datum % value == 0


def keys_contain(trial_dict, key):
    return key in trial_dict.keys()


def keys_contain_any_of(trial_dict, *keys):
    return any(k in trial_dict.keys() for k in keys)


def keys_contain_all_of(trial_dict, *keys):
    return all(k in trial_dict.keys() for k in keys)


def keys_contain_N_of(trial_dict, N, keys):
    return sum(k in trial_dict.keys() for k in keys) == N


def keys_contain_at_least_N_of(trial_dict, N, keys):
    return sum(k in trial_dict.keys() for k in keys) >= N


def keys_contain_at_most_N_of(trial_dict, N, keys):
    return sum(k in trial_dict.keys() for k in keys) <= N


def keys_contain_one_of(trial_dict, *keys):
    return keys_contain_N_of(trial_dict, 1, keys)


def keys_contain_at_least_one_of(trial_dict, keys):
    return keys_contain_at_least_N_of(trial_dict, 1, keys)


def keys_contain_at_most_one_of(trial_dict, keys):
    return keys_contain_at_most_N_of(trial_dict, 1, keys)


def keys_equal_to(trial_dict, *keys):
    return set(trial_dict.keys()) == set(keys)


def items_contain(trial_dict, **items):
    for k, v in items.items():
        try:
            if trial_dict[k] != v:
                return False
        except KeyError:
            return False
    return True


def allowed_keys(trial_dict, *keys):
    return not (set(trial_dict.keys()) - set(keys))


def required_keys(trial_dict, *keys):
    return not (set(keys) - set(trial_dict.keys()))


def forbidden_keys(trial_dict, *keys):
    return not (set(keys) & set(trial_dict.keys()))


def equal_to_approx(trial_datum, value, tolerance):
    return abs(trial_datum - value) < tolerance


def truthy(trial_datum):
    return not not trial_datum


def falsy(trial_datum):
    return not trial_datum


def null(trial_data):
    return True


def is_instance(trial_datum, *classes):
    return isinstance(trial_datum, classes)



import enum
import operator
import pathlib
from subprocess import call
import warnings


import valida.data
import valida.datapath
from valida import callables as call_funcs
from valida.errors import (
    InvalidCallable,
    MalformedConditionLikeSpec,
    MalformedDataPathSpec,
)
from valida.utils import (
    classproperty,
    get_func_args_by_kind,
    null_condition_binary_check,
)


class PreparedConditionCallable:
    def __init__(self, func, *args, **kwargs):
        self._func = func
        self._args = args
        self._kwargs = kwargs

    def __call__(self, trial_datum, source_data=None):
        res_args, res_kwargs = self._get_resolved_data_path_args(source_data)
        return self.func(trial_datum, *res_args, **res_kwargs)

    def _get_resolved_data_path_args(self, source_data):
        if not source_data:
            return self.args, self.kwargs

        resolved_args = []
        for arg in self.args:
            if isinstance(arg, valida.datapath.DataPath):
                arg = arg.get_data(source_data, return_paths=False)
            resolved_args.append(arg)

        resolved_kwargs = {}
        for k, v in self.kwargs.items():
            if isinstance(v, valida.datapath.DataPath):
                v = v.get_data(source_data, return_paths=False)
            resolved_kwargs[k] = v

        return tuple(resolved_args), resolved_kwargs

    @property
    def func(self):
        return self._func

    @property
    def name(self):
        return self.func.__name__

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs


class GeneralCallables:
    @classmethod
    def equal_to(cls, value):
        return cls(call_funcs.equal_to, value=value)

    @classmethod
    def not_equal_to(cls, value):
        return cls(call_funcs.not_equal_to, value=value)

    @classmethod
    def less_than(cls, value):
        return cls(call_funcs.less_than, value=value)

    @classmethod
    def greater_than(cls, value):
        return cls(call_funcs.greater_than, value=value)

    @classmethod
    def less_than_or_equal_to(cls, value):
        return cls(call_funcs.less_than_or_equal_to, value=value)

    @classmethod
    def greater_than_or_equal_to(cls, value):
        return cls(call_funcs.greater_than_or_equal_to, value=value)

    @classmethod
    def in_(cls, value):
        return cls(call_funcs.in_, value=value)

    @classmethod
    def not_in(cls, value):
        return cls(call_funcs.not_in, value=value)

    @classmethod
    def in_range(cls, lower, upper):
        return cls(call_funcs.in_range, lower=lower, upper=upper)

    @classmethod
    def not_in_range(cls, value):
        return cls(call_funcs.not_in_range, value=value)

    @classmethod
    def equal_to_approx(cls, value, tolerance=1e-8):
        return cls(call_funcs.equal_to_approx, value=value, tolerance=tolerance)

    @classmethod
    def factor_of(cls, value):
        return cls(call_funcs.factor_of, value)

    @classmethod
    def has_factor(cls, value):
        return cls(call_funcs.has_factor, value)

    @classmethod
    def truthy(cls):
        return cls(call_funcs.truthy)

    @classmethod
    def falsy(cls):
        return cls(call_funcs.falsy)

    @classmethod
    def null(cls):
        return cls(call_funcs.null)

    @classmethod
    def is_instance(cls, *classes):
        return cls(call_funcs.is_instance, *classes)

    # Aliases for convenience:
    eq = equal_to
    lt = less_than
    gt = greater_than
    lte = less_than_or_equal_to
    gte = greater_than_or_equal_to

    OP_SYMBOL_MAP = {
        "==": equal_to,
        "<": less_than,
        ">": greater_than,
        "<=": less_than_or_equal_to,
        ">=": greater_than_or_equal_to,
    }


class MapCallables:
    @classmethod
    def keys_contain(cls, key):
        return cls(call_funcs.keys_contain, key=key)

    @classmethod
    def keys_contain_any_of(cls, *keys):
        return cls(call_funcs.keys_contain_any_of, *keys)

    @classmethod
    def keys_contain_all_of(cls, *keys):
        return cls(call_funcs.keys_contain_all_of, *keys)

    @classmethod
    def keys_contain_N_of(cls, N, keys):
        return cls(call_funcs.keys_contain_N_of, N=N, keys=keys)

    @classmethod
    def keys_contain_at_least_N_of(cls, N, keys):
        return cls(call_funcs.keys_contain_at_least_N_of, N=N, keys=keys)

    @classmethod
    def keys_contain_at_most_N_of(cls, N, keys):
        return cls(call_funcs.keys_contain_at_most_N_of, N=N, keys=keys)

    @classmethod
    def keys_contain_one_of(cls, *keys):
        return cls(call_funcs.keys_contain_one_of, *keys)

    @classmethod
    def keys_contain_at_least_one_of(cls, keys):
        return cls(call_funcs.keys_contain_at_least_one_of, keys=keys)

    @classmethod
    def keys_contain_at_most_one_of(cls, keys):
        return cls(call_funcs.keys_contain_at_most_one_of, keys=keys)

    @classmethod
    def keys_equal_to(cls, *keys):
        return cls(call_funcs.keys_equal_to, *keys)

    @classmethod
    def items_contain(cls, **items):
        return cls(call_funcs.items_contain, **items)

    @classmethod
    def allowed_keys(cls, *keys):
        return cls(call_funcs.allowed_keys, *keys)

    @classmethod
    def required_keys(cls, *keys):
        return cls(call_funcs.required_keys, *keys)

    @classmethod
    def forbidden_keys(cls, *keys):
        return cls(call_funcs.forbidden_keys, *keys)


class AllCallables(GeneralCallables, MapCallables):
    pass


class ConditionLike:
    def __or__(self, other):
        return ConditionOr(self, other)

    def __and__(self, other):
        return ConditionAnd(self, other)

    def __xor__(self, other):
        return ConditionXor(self, other)

    def is_like(self, cls):
        return all(isinstance(i, cls) for i in self.flatten()[0])

    @property
    def is_null(self):
        return isinstance(self, NullCondition)

    @property
    def is_key_like(self):
        return self.is_like(KeyLike)

    @property
    def is_index_like(self):
        return self.is_like(IndexLike)

    @property
    def is_value_like(self):
        return self.is_like(ValueLike)

    def filter(self, data, data_has_paths=False, source_data=None):
        if not isinstance(data, valida.data.Data):
            data = valida.data.Data(data)
        return self._filter(
            data, data_has_paths=data_has_paths, source_data=source_data
        )

    def test(self, datum):
        return self.filter([datum]).result[0]

    def test_all(self, data):
        return all(self.filter(data).result)

    def flatten(self):
        """Get a flattened list of all conditions."""
        all_cnds = []
        all_ops = []

        try:
            for idx, cnd_i in enumerate(self.children):
                flatten_i = cnd_i.flatten()
                all_cnds.extend(flatten_i[0])
                all_ops.extend(flatten_i[1])
                if idx == 0:
                    all_ops.append(self.FLATTEN_SYMBOL)

        except AttributeError:
            all_cnds.append(self)

        return all_cnds, all_ops

    @staticmethod
    def from_spec(spec):

        if not spec:
            return NullCondition()
        elif not isinstance(spec, dict):
            raise TypeError("`spec` must be a dict with a single item.")

        BINARY_OPS = {
            "and": ConditionAnd,
            "or": ConditionOr,
            "xor": ConditionXor,
        }
        CONDITION_DATUM_TYPES = {
            "value": Value,
            "key": Key,
            "index": Index,
        }
        CALLABLE_LOOKUP = {
            "in": "in_",
        }
        PRE_PROC_LOOKUP = {
            "type": "dtype",
            "dtype": "dtype",
            "length": "length",
            "len": "length",
        }
        DTYPE_LOOKUP = {
            "int": int,
            "float": float,
            "str": str,
            "list": list,
            "dict": dict,
            "map": dict,
            "bool": bool,
            "path": pathlib.Path,
            int: int,
            float: float,
            str: str,
            list: list,
            dict: dict,
            bool: bool,
            pathlib.Path: pathlib.Path,
        }
        ALL_PRE_PROCS = list(PRE_PROC_LOOKUP.keys())

        if len(spec) > 1:
            raise MalformedConditionLikeSpec(
                f"A condition-like should be specified with exactly one "
                f"specification key (but found keys: {list(spec.keys())}). Allowed specifications "
                f"keys are the binary operators: {list(BINARY_OPS.keys())!r}, and the "
                f"condition-specifications, which are strings that start with one of the "
                f"condition datum types: {list(CONDITION_DATUM_TYPES.keys())!r}."
            )
        else:
            spec_key, spec_val = next(iter(spec.items()))  # single-item dict

        spec_key_split = [i.lower() for i in spec_key.split(".")]
        spec_key_split_len = len(spec_key_split)

        if spec_key in BINARY_OPS:
            cls = BINARY_OPS[spec_key]

            if not isinstance(spec_val, (list, tuple)):
                raise MalformedConditionLikeSpec(
                    f"A condition binary operator must be associated with a "
                    f"*list* of condition-likes, but value is not list-like, with type "
                    f"{type(spec_val)!r}."
                )

            condition_like = NullCondition()
            for i in spec_val:
                i_obj = ConditionLike.from_spec(i)
                condition_like = cls(condition_like, i_obj)

        elif spec_key_split[0] in CONDITION_DATUM_TYPES:

            if (
                spec_key_split_len not in [2, 3]
                or spec_key_split_len == 2
                and spec_key_split[-1] in ALL_PRE_PROCS
            ):
                raise MalformedConditionLikeSpec(
                    f"Condition specification must be a dot-delimited string with either "
                    f"two tokens: <datum type>.<callable name>; or three tokens "
                    f"<datum type>.<pre-processor name>.<callable name>, but specification "
                    f"was given as: {spec_key!r}."
                )

            condition_type_str = spec_key_split[0]
            cls = CONDITION_DATUM_TYPES[condition_type_str]
            pre_proc_str = None

            if spec_key_split_len == 3:

                try:
                    pre_proc_str = spec_key_split[1]
                    pre_proc_str = PRE_PROC_LOOKUP.get(pre_proc_str, pre_proc_str)
                    if pre_proc_str == "dtype":
                        try:
                            # convert strings to types
                            if isinstance(spec_val, list):
                                spec_val = [
                                    DTYPE_LOOKUP[i.lower() if isinstance(i, str) else i]
                                    for i in spec_val
                                ]
                            else:
                                spec_val = DTYPE_LOOKUP[
                                    spec_val.lower()
                                    if isinstance(spec_val, str)
                                    else spec_val
                                ]
                        except KeyError:
                            raise MalformedConditionLikeSpec(
                                f"Data type {spec_val!r} is not understood. Available data "
                                f"types are: {list(DTYPE_LOOKUP.keys())!r}."
                            )

                    cls = getattr(cls, pre_proc_str)

                except AttributeError:
                    raise MalformedConditionLikeSpec(
                        f"Condition pre-processor {pre_proc_str!r} not understood. "
                        f'Available pre-processors are "length" and "type"/"dtype", but '
                        f"not all pre-processors are applicable to all condition types."
                    )

            cond_call_str = spec_key_split[-1]
            cond_call_str = CALLABLE_LOOKUP.get(cond_call_str, cond_call_str)
            # special case:
            if cond_call_str == "is_instance":
                try:
                    # convert strings to types
                    if isinstance(spec_val, list):
                        spec_val = [
                            DTYPE_LOOKUP[i.lower() if isinstance(i, str) else i]
                            for i in spec_val
                        ]
                    else:
                        spec_val = DTYPE_LOOKUP[
                            spec_val.lower() if isinstance(spec_val, str) else spec_val
                        ]
                except KeyError:
                    raise MalformedConditionLikeSpec(
                        f"Data type {spec_val!r} is not understood. Available data "
                        f"types are: {list(DTYPE_LOOKUP.keys())!r}."
                    )

            try:
                cond_method = getattr(cls, cond_call_str)
            except AttributeError:
                msg = (
                    f'Condition callable "{cond_call_str}" is not known or not '
                    f'compatible with specified condition type "{condition_type_str}"'
                )
                if pre_proc_str:
                    msg += f' and condition pre-processor "{pre_proc_str}"'
                msg += "."
                raise MalformedConditionLikeSpec(msg)

            func_args = get_func_args_by_kind(cond_method)

            # Coerce arguments to `DataPath`s where specified as such:
            if isinstance(spec_val, dict):
                try:
                    spec_val = valida.datapath.DataPath.from_spec(spec_val)
                except MalformedDataPathSpec:
                    # Check values for DataPath specs:
                    for k, v in spec_val.items():
                        try:
                            spec_val[k] = valida.datapath.DataPath.from_spec(v)
                        except MalformedDataPathSpec:
                            pass
            elif isinstance(spec_val, (list, tuple)):
                # Check items for DataPath specs:
                for idx, v in enumerate(spec_val):
                    try:
                        spec_val[idx] = valida.datapath.DataPath.from_spec(v)
                    except MalformedDataPathSpec:
                        pass

            if not any(
                func_args[i]
                for i in ("POSITIONAL_OR_KEYWORD", "VAR_POSITIONAL", "VAR_KEYWORD")
            ):
                # no arguments, check value given is None and invoke:
                if spec_val is not None:
                    warnings.warn(
                        f'Condition callable "{cond_method}" takes no argument, but an '
                        f'argument was supplied: "{spec_val}". This will be ignored.'
                    )
                condition = cond_method()

            elif len(func_args["POSITIONAL_OR_KEYWORD"]) == 1 and not any(
                func_args[i] for i in ("VAR_POSITIONAL", "VAR_KEYWORD")
            ):
                # exactly one single positional-or-keyword; pass as a positional:
                condition = cond_method(spec_val)

            elif len(func_args["POSITIONAL_OR_KEYWORD"]) > 1 and not any(
                func_args[i] for i in ("VAR_POSITIONAL", "VAR_KEYWORD")
            ):
                # More than one positional-or-keyword, and no other kinds:
                if isinstance(spec_val, dict):
                    condition = cond_method(**spec_val)
                elif isinstance(spec_val, (list, tuple)):
                    condition = cond_method(*spec_val)
                else:
                    raise MalformedConditionLikeSpec(
                        f"Condition callable {cond_method} accepts multuple positional-or-"
                        f"keyword arguments, and so must be parametrised with a dict or "
                        f"list/tuple of values, but the following was supplied: {spec_val!r}"
                    )

            elif len(func_args["VAR_POSITIONAL"]) == 1 and not any(
                func_args[i] for i in ("POSITIONAL_OR_KEYWORD", "VAR_KEYWORD")
            ):
                if not isinstance(spec_val, list):
                    raise MalformedConditionLikeSpec(
                        f'Condition callable "{cond_method}" accepts a variable-positional '
                        f"argument, and so must be parametrised with a list of values, "
                        f'but the following was supplied: "{spec_val}".'
                    )
                condition = cond_method(*spec_val)

            elif len(func_args["VAR_KEYWORD"]) == 1 and not func_args["VAR_POSITIONAL"]:
                # zero or more positional-or-keyword args and a variable-keyword arg:
                if not isinstance(spec_val, dict):
                    raise MalformedConditionLikeSpec(
                        f'Condition callable "{cond_method}" accepts a variable-keyword '
                        f"argument and zero or more positional-or-keyword arguments, and "
                        f"so must be parametrised with a dict of values, but the following "
                        f'was supplied: "{spec_val}".'
                    )
                condition = cond_method(**spec_val)

            else:
                raise MalformedConditionLikeSpec(
                    f"Condition callable arguments {spec_val!r} with type {type(spec_val)} "
                    f"not allowed."
                )
            condition_like = condition

        else:
            raise MalformedConditionLikeSpec(
                f"Condition-like specification keys not understood: {list(spec.keys())}). "
                f"Allowed specifications keys are the binary operators: "
                f"{list(BINARY_OPS.keys())!r}, and the condition-specifications, which are "
                f"strings that start with one of the condition datum types: "
                f"{list(CONDITION_DATUM_TYPES.keys())!r}."
            )

        return condition_like

    @classmethod
    def from_json_like(cls, json_like, *args, **kwargs):
        return cls.from_spec(json_like)

class Condition(ConditionLike):

    PRE_PROCESSOR = None

    def __init__(self, callable, *args, **kwargs):
        """
        Parameters
        ----------
        callable : Callable
            A callable whose first argument is the trial datum that is to be tested.
        args : tuple
            Callable positional arguments.
        kwargs : dict
            Callable keyword arguments.

        """

        self.callable = PreparedConditionCallable(callable, *args, **kwargs)

    def __repr__(self):

        out = f"{self.__class__.__name__}.{self.callable.name}"
        args = [f"{v!r}" for v in self.callable.args] + [
            f"{k}={v!r}" for k, v in self.callable.kwargs.items()
        ]
        out += "(" + ", ".join(args) + ")"

        return out

    def __eq__(self, other):
        if type(other) is type(self) and self._members() == other._members():
            return True
        return False

    def _members(self):
        """Return data used in __eq__"""
        return (
            self.callable.name,
            self.callable.args,
            self.callable.kwargs,
        )

    def _filter(self, data, data_has_paths=False, source_data=None):

        processed = []
        pre_processor_error = []
        callable_error = []
        callable_false = []
        for datum in getattr(data, self.DATUM_TYPE.value)():

            if data_has_paths:
                datum, _ = datum

            try:
                processed_i = self.PRE_PROCESSOR(datum) if self.PRE_PROCESSOR else datum
                is_valid_i = True
                pre_processor_error_i = False
            except TypeError:
                processed_i = None
                is_valid_i = False
                pre_processor_error_i = True

            callable_error_i = False
            callable_false_i = False
            if is_valid_i:

                try:
                    result_i = self.callable(processed_i, source_data=source_data)

                    if not isinstance(result_i, bool):
                        raise InvalidCallable(
                            f"Callable {self.callable} did not return a boolean."
                        )
                    if not result_i:
                        callable_false_i = True

                except (TypeError, AttributeError):
                    callable_error_i = True

            pre_processor_error.append(pre_processor_error_i)
            callable_error.append(callable_error_i)
            callable_false.append(callable_false_i)
            processed.append(processed_i)

        return valida.data.FilteredData(
            self,
            data,
            processed,
            pre_processor_error,
            callable_error,
            callable_false,
            data_has_paths=data_has_paths,
        )

    def to_json_like(self, *args, **kwargs):
        # need to return a single-item dict that can be passed to `from_spec` for
        # round-tripping.

        INV_DTYPE_LOOKUP = {
            int: "int",
            float: "float",
            str: "str",
            list: "list",
            dict: "dict",
            bool: "bool",
            pathlib.Path: "path",
        }

        # TODO: doesn't work for BinaryOps?
        key = f"{self.js_like_label}.{self.callable.func.__name__}"

        if len(self.callable.kwargs) == 1:
            _, val = next(iter(self.callable.kwargs.items()))
            if "dtype" in key:
                val = INV_DTYPE_LOOKUP[val]

        out = {key: val}
        if "shared_data" in kwargs:
            return out, kwargs["shared_data"]
        else:
            return out


class FilterDatumType(enum.Enum):

    KEYS = "keys"
    VALUES = "values"


class NullCondition(Condition):
    """Class to represent a null condition that all data satisfies."""

    DATUM_TYPE = FilterDatumType.VALUES

    def __init__(self, *args, **kwargs):
        super().__init__(call_funcs.null)

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def to_json_like(self, *args, **kwargs):
        if "shared_data" in kwargs:
            return {}, kwargs["shared_data"]
        else:
            return {}


class ConditionBinaryOp(ConditionLike):
    def __new__(cls, *conditions):
        """If one of the conditions is a NullCondition, then abort object construction,
        and just return the non-null condition."""
        return null_condition_binary_check(*conditions) or super().__new__(cls)

    def __init__(self, *conditions):

        super().__init__()

        self.children = conditions

        # Nonsensical to combine key-like and index-like conditions:
        flattened_conds = self.flatten()[0]
        num_key_likes = sum(isinstance(i, KeyLike) for i in flattened_conds)
        num_index_likes = sum(isinstance(i, IndexLike) for i in flattened_conds)
        if num_key_likes > 0 and num_index_likes > 0:
            raise TypeError("Cannot combine `Key` and `Index` conditions.")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.children[0]}, {self.children[1]})"

    def __eq__(self, other):
        if type(self) is type(other) and (
            (
                self.children[0] == other.children[0]
                and self.children[1] == other.children[1]
            )
            or (
                self.children[0] == other.children[1]
                and self.children[1] == other.children[0]
            )
        ):
            return True
        return False

    def _filter(self, data, binary_op, data_has_paths=False, source_data=None):

        if data_has_paths:
            # Data paths will be removed on the first child _filter:
            data_has_paths = [True, False]
        else:
            data_has_paths = [False, False]

        return binary_op(
            *(
                i._filter(data, data_has_paths[idx], source_data=source_data)
                for idx, i in enumerate(self.children)
            )
        )


class ConditionAnd(ConditionBinaryOp):
    FLATTEN_SYMBOL = "and"

    def _filter(self, data, data_has_paths=False, source_data=None):
        return super()._filter(
            data, operator.and_, data_has_paths=data_has_paths, source_data=source_data
        )


class ConditionOr(ConditionBinaryOp):
    FLATTEN_SYMBOL = "or"

    def _filter(self, data, data_has_paths=False, source_data=None):
        return super()._filter(
            data, operator.or_, data_has_paths=data_has_paths, source_data=source_data
        )


class ConditionXor(ConditionBinaryOp):
    FLATTEN_SYMBOL = "xor"

    def _filter(self, data, data_has_paths=False, source_data=None):
        return super()._filter(
            data, operator.xor, data_has_paths=data_has_paths, source_data=source_data
        )


class LengthPreProcessor:
    PRE_PROCESSOR = len


class DataTypePreProcessor:
    PRE_PROCESSOR = type


class ValueLike(Condition):
    DATUM_TYPE = FilterDatumType.VALUES


class KeyLike(Condition):
    DATUM_TYPE = FilterDatumType.KEYS

    def filter(self, data, data_has_paths=False, source_data=None):

        if (isinstance(data, valida.data.Data) and data.is_list) or not isinstance(
            data, (valida.data.Data, dict)
        ):
            raise TypeError("`Key` condition can only filter a mapping (i.e. a dict).")

        return super().filter(
            data, data_has_paths=data_has_paths, source_data=source_data
        )

    def test(self, datum):
        """For testing a single-item mapping."""
        if len(datum) != 1:
            raise TypeError("Test can only be used to test a single-item mapping.")
        return self.filter(datum).result[0]


class IndexLike(Condition):
    DATUM_TYPE = FilterDatumType.KEYS

    def filter(self, data, data_has_paths=False, source_data=None):

        if (isinstance(data, valida.data.Data) and not data.is_list) or not isinstance(
            data, (valida.data.Data, list)
        ):
            raise TypeError("`Index` condition can only filter a list.")
        return super().filter(
            data, data_has_paths=data_has_paths, source_data=source_data
        )

    def test(self, datum):
        """Not allowed, since the index of a list item is unknown to the item itself."""
        raise NotImplementedError


class Value(ValueLike, AllCallables):
    js_like_label = "value"

    @classproperty
    def length(cls):
        return ValueLength

    @classproperty
    def dtype(cls):
        return ValueDataType


class ValueLength(LengthPreProcessor, ValueLike, GeneralCallables):
    js_like_label = "value.length"


class ValueDataType(DataTypePreProcessor, ValueLike, GeneralCallables):
    js_like_label = "value.dtype"


class Key(KeyLike, AllCallables):
    js_like_label = "key"

    @classproperty
    def length(cls):
        return KeyLength

    @classproperty
    def dtype(cls):
        return KeyDataType


class KeyLength(LengthPreProcessor, KeyLike, GeneralCallables):
    js_like_label = "key.length"


class KeyDataType(DataTypePreProcessor, KeyLike, GeneralCallables):
    js_like_label = "key.dtype"


class Index(IndexLike, GeneralCallables):
    js_like_label = "index"



__version__ = "0.5.1"

from valida.conditions import Value, Key, Index
from valida.data import Data
from valida.datapath import DataPath
from valida.rules import Rule
from valida.schema import Schema



import operator

import valida.datapath


def set_datum(data, data_path, datum):

    for part in data_path.parts[:-1]:
        idx = part.condition.callable.kwargs["value"]
        data = data[idx]

    idx = data_path.parts[-1].condition.callable.kwargs["value"]
    data[idx] = datum


class Data:
    def __init__(self, data):

        if not isinstance(data, (list, dict)) or not data:
            raise TypeError(f"Data is not filterable: {data!r}.")

        try:
            keys, values = zip(*data.items())
            is_list = False
        except AttributeError:
            keys, values = range(len(data)), data
            is_list = True

        self._keys = keys
        self._values = values
        self._is_list = is_list

    def get_original(self):
        if self.is_list:
            return list(self._values)
        else:
            return {k: v for k, v in zip(self._keys, self._values)}

    def __len__(self):
        return len(self.keys())

    def __eq__(self, other):
        if not isinstance(other, Data):
            return False
        if self.is_list != other.is_list:
            return False
        if self.keys() != other.keys():
            return False
        if self.values() != other.values():
            return False
        return True

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(keys={self.keys()!r}, values={self.values()!r})"
        )

    def __iter__(self):
        for idx in range(len(self)):
            yield self.keys()[idx]

    def __getitem__(self, idx):
        # Note we always expect a values *index* regardless of `self.is_list`
        return self.values()[idx]

    def extract_paths(self):
        values, concrete_paths = list(zip(*self.values()))
        self._values = list(values)
        return concrete_paths

    @property
    def original(self):
        if self.is_list:
            return self.values()
        else:
            return {i: j for i, j in zip(self.keys(), self.values())}

    @property
    def is_list(self):
        return self._is_list

    def keys(self):
        return self._keys

    def values(self):
        return self._values

    def items(self):
        return iter(zip(self.keys(), self.values()))

    def filter(self, condition_like):
        return condition_like.filter(self)

    def get(self, *path_parts, return_paths=False):
        """Get a subset of data from this data.

        Parameters
        ----------
        path_parts : DataPath or tuple of path parts

        Examples
        --------

        For "concrete paths" whose parts are dict keys or list indices, a single
        value is returned:

        >>> data = Data({'a': {'b': [1, 2.5, 3]}})
        >>> data.get('a')
        >>> {'b', [1, 2.5, 3]}

        >>> data.get('a', 'b')
        >>> [1, 2.5, 3]

        >>> data.get('a', 'b', 1)
        >>> 2.5

        For "non-concrete paths" whose parts include `MapValue` and `ListValue`
        objects, a dict or list of values is returned:

        >>> data_path = DataPath(MapValue('a'), 'b')
        >>> data.get(data_path)
        >>> [[1, 2.5, 3]]

        >>> data_path = DataPath('a', 'b', ListValue(value=Value.equal_to(1)))
        >>> data.get(data_path)
        >>> [1]

        >>> data_path = DataPath('a', 'b', ListValue(index=Index.in_([1, 2])))
        >>> data.get(data_path)
        >>> [2.5, 3]

        """
        if not path_parts:
            data_path = valida.datapath.DataPath()
        elif isinstance(path_parts[0], valida.datapath.DataPath):
            data_path = path_parts[0]
        else:
            data_path = valida.datapath.DataPath(*path_parts)

        return data_path.get_data(self, return_paths=return_paths)

    def set(self, path, datum):
        """Return a copy, with a new value set at the supplied path."""
        if not path.is_concrete:
            raise TypeError("Cannot set data at non-concrete path.")

        data = self.get_original()
        set_datum(data, path, datum)
        return Data(data)


class FilteredDataLike:
    @property
    def data(self):
        return [i for idx, i in enumerate(self.source.values()) if self.result[idx]]

    @property
    def keys(self):
        """These are list indices for the case of a list-like data."""
        return [k for idx, k in enumerate(self.source.keys()) if self.result[idx]]

    def __iter__(self):
        for idx, _ in enumerate(self.result):
            yield FilteredDataItem(self, idx)

    def __getitem__(self, index):
        return FilteredDataItem(self, index)

    def __and__(self, other):
        return FilteredDataAnd(self, other)

    def __or__(self, other):
        return FilteredDataOr(self, other)

    def __xor__(self, other):
        return FilteredDataXor(self, other)

    @property
    def failure_indices(self):
        return [idx for idx, i in enumerate(self.result) if not i]

    def get_all_failures(self):
        failures = []
        for fail_idx in self.failure_indices:
            failures.append(self.get_failure_by_index(fail_idx))
        return failures

    def get_failure_by_index(self, idx):

        if self.result[idx]:
            return None

        else:

            truth_table_error_messages = (
                (
                    self.truth_table_pre_processor_error,
                    "Condition pre-processor raised an exception: `{}`.",
                ),
                (
                    self.truth_table_callable_error,
                    "Condition callable raised an exception: `{}`.",
                ),
                (
                    self.truth_table_callable_false,
                    "Condition callable returned False: `{}`.",
                ),
            )

            failure = []
            for cnd_idx, (cnd_name, _) in enumerate(self.truth_table):

                if cnd_name in ("and", "or"):
                    # xor condition returns False if both child conditions are True
                    continue

                for truth_table, err_msg in truth_table_error_messages:
                    if truth_table[cnd_idx][1][idx]:
                        failure.append(err_msg.format(cnd_name))
                        break

        return tuple(failure)


class FilteredData(FilteredDataLike):
    def __init__(
        self,
        condition,
        source: Data,
        processed,
        pre_processor_error,
        callable_error,
        callable_false,
        data_has_paths=False,
    ):

        self.condition = condition
        self.processed = processed

        self.pre_processor_error = pre_processor_error
        self.callable_error = callable_error
        self.callable_false = callable_false

        self.result = [
            False if i else (False if j else (False if k else True))
            for i, j, k in zip(
                self.pre_processor_error, self.callable_error, self.callable_false
            )
        ]

        self.concrete_paths = None
        self.source = source

        if data_has_paths:
            self.concrete_paths = self.source.extract_paths()

    def __eq__(self, other):
        if type(self) == type(other):
            if self.condition == other.condition:
                if self.source == other.source:
                    return True
        return False

    def _get_truth_table(self, attribute):
        return [(repr(self.condition), getattr(self, attribute))]

    @property
    def truth_table(self):
        return self.truth_table_result

    @property
    def truth_table_result(self):
        return self._get_truth_table("result")

    @property
    def truth_table_pre_processor_error(self):
        return self._get_truth_table("pre_processor_error")

    @property
    def truth_table_callable_error(self):
        return self._get_truth_table("callable_error")

    @property
    def truth_table_callable_false(self):
        return self._get_truth_table("callable_false")


class FilteredDataItem:
    def __init__(self, filtered_data_like, index):
        self.filtered_data_like = filtered_data_like
        self.index = index

        self.source = self.filtered_data_like.source[self.index]
        self.result = self.filtered_data_like.result[self.index]
        self.concrete_path = self.filtered_data_like.concrete_paths[self.index]

    def get_failure(self):
        return self.filtered_data_like.get_failure_by_index(self.index)


class FilteredDataBinaryOp(FilteredDataLike):
    def __init__(self, fd1, fd2, binary_op):

        if fd1.source is not fd2.source:
            raise RuntimeError(
                f"The two Data `source` objects in a `{self.__class__.__name__}` must be "
                f"the same object."
            )

        self.source = fd1.source
        self.children = [fd1, fd2]
        self.result = [binary_op(i, j) for i, j in zip(fd1.result, fd2.result)]
        self.concrete_paths = fd1.concrete_paths

        self.pre_processor_error = [
            any((i, j))
            for i, j in zip(fd1.pre_processor_error, fd2.pre_processor_error)
        ]
        self.callable_error = [
            any((i, j)) for i, j in zip(fd1.callable_error, fd2.callable_error)
        ]
        self.callable_false = [not i for i in self.result]

    def _get_truth_table(self, attribute):
        return (
            self.children[0]._get_truth_table(attribute)
            + self.children[1]._get_truth_table(attribute)
            + [(self.TRUTH_TABLE_SYMBOL, getattr(self, attribute))]
        )

    @property
    def truth_table(self):
        return self.truth_table_result

    @property
    def truth_table_result(self):
        return self._get_truth_table("result")

    @property
    def truth_table_pre_processor_error(self):
        return self._get_truth_table("pre_processor_error")

    @property
    def truth_table_callable_error(self):
        return self._get_truth_table("callable_error")

    @property
    def truth_table_callable_false(self):
        return self._get_truth_table("callable_false")


class FilteredDataAnd(FilteredDataBinaryOp):
    TRUTH_TABLE_SYMBOL = "and"

    def __init__(self, fd1, fd2):
        super().__init__(fd1, fd2, operator.and_)


class FilteredDataOr(FilteredDataBinaryOp):
    TRUTH_TABLE_SYMBOL = "or"

    def __init__(self, fd1, fd2):
        super().__init__(fd1, fd2, operator.or_)


class FilteredDataXor(FilteredDataBinaryOp):
    TRUTH_TABLE_SYMBOL = "xor"

    def __init__(self, fd1, fd2):
        super().__init__(fd1, fd2, operator.xor)



import copy
from valida.conditions import ConditionLike
from valida.casting import CAST_DTYPE_LOOKUP, CAST_LOOKUP
from valida.data import Data, set_datum
from valida.datapath import DataPath
from valida.errors import MalformedRuleSpec


class Rule:
    def __init__(self, path, condition, cast=None):

        if not isinstance(path, DataPath):
            path = DataPath(*path)

        self.path = path
        self.condition = condition
        self.cast = cast

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"path={self.path!r}, condition={self.condition!r}, cast={self.cast!r}"
            f")"
        )

    def __eq__(self, other):
        if type(other) == type(self):
            if (
                other.path == self.path
                and other.condition == self.condition
                and other.cast == self.cast
            ):
                return True
        return False

    @classmethod
    def from_spec(cls, spec):
        path = DataPath.from_part_specs(*spec["path"])
        cond = ConditionLike.from_spec(spec["condition"])
        cast = spec.get("cast")
        for cast_from in list((cast or {}).keys()):

            cast_to = cast.pop(cast_from)
            try:
                cast_from = CAST_DTYPE_LOOKUP[cast_from]
            except KeyError:
                raise MalformedRuleSpec(f"Unknown cast_from type: {cast_from!r}")

            try:
                cast_to = CAST_DTYPE_LOOKUP[cast_to]
            except KeyError:
                raise MalformedRuleSpec(f"Unknown cast_to type: {cast_to!r}")

            try:
                cast[cast_from] = CAST_LOOKUP[(cast_from, cast_to)]
            except KeyError:
                raise MalformedRuleSpec(
                    f"Unsupported cast from type {cast_from!r} to type {cast_to!r}."
                )

        return cls(path=path, condition=cond, cast=cast)

    @classmethod
    def from_json_like(cls, json_like, *args, **kwargs):
        return cls.from_spec(json_like)

    def to_json_like(self, *args, **kwargs):
        out = {
            "condition": self.condition.to_json_like(),
            "cast": self.cast,
            "path": self.path.to_json_like(),
        }
        if "shared_data" in kwargs:
            return out, kwargs["shared_data"]
        else:
            return out

    def test(self, data, _data_copy=None):

        if not isinstance(data, Data):
            data = Data(data)

        if not self.cast:
            data_copy = data
        else:
            data_copy = _data_copy or copy.deepcopy(data.get_original())
            sub_data = self.path.get_data(data, return_paths=True)
            path_exists = sub_data not in [None, []]
            if self.path.is_concrete:
                sub_data = [sub_data]
            if path_exists:
                for datum, datum_path in sub_data:
                    for k, v in self.cast.items():
                        if isinstance(datum, k):
                            try:
                                datum = v(datum)
                                break
                            except TypeError:
                                pass
                    datum_path = DataPath(*datum_path)
                    set_datum(data_copy, datum_path, datum)

        return RuleTest(self, data_copy)


class RuleTestFailureItem:
    def __init__(self, rule_test, index, value, path, reasons):

        self.rule_test = rule_test
        self.index = index
        self.value = value
        self.path = path
        self.reasons = reasons

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"value={self.value!r}, path={self.path!r}), reasons={self.reasons!r}"
            f")"
        )


class RuleTest:
    def __init__(self, rule, data):

        if not isinstance(data, Data):
            data = Data(data)

        self.rule = rule
        self.data = data

        self._tested = False  # assigned by `_test()` - True if the rule path existed
        self._is_valid = None  # assigned by `_test()`
        self._failures = None  # assigned by `_test()`

        self._test()

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"is_valid={self.is_valid!r}, num_failures={self.num_failures!r}"
            f")"
        )

    def __eq__(self, other):
        if type(other) == type(self):
            if other.rule == self.rule and other.data is self.data:
                return True
        return False

    @property
    def is_valid(self):
        return self._is_valid

    @property
    def tested(self):
        return self._tested

    @property
    def num_failures(self):
        return len(self.failures)

    @property
    def failures(self):
        return self._failures

    def get_failures_string(self):
        out = ""
        if not self.failures:
            out += "Rule test is valid.\n"
        for fail in self.failures:
            out += f"Path: {fail.path!r}\nValue: {fail.value!r}\nReasons:\n"
            for reason in fail.reasons:
                out += " " + reason + "\n"
        return out

    def print_failures(self):
        print(self.get_failures_string())

    def _test(self):

        sub_data = self.rule.path.get_data(self.data, return_paths=True)
        path_exists = sub_data not in [None, []]

        if self.rule.path.is_concrete:
            sub_data = [sub_data]

        self.sub_data = sub_data

        failures = []
        if path_exists:
            filtered_data = self.rule.condition.filter(
                sub_data,
                data_has_paths=True,
                source_data=self.data,
            )
            if all(filtered_data.result):
                self._is_valid = True

            else:
                self._is_valid = False
                for f_item in filtered_data:
                    if not f_item.result:
                        failure_item = RuleTestFailureItem(
                            rule_test=self,
                            index=f_item.index,
                            value=f_item.source,
                            path=f_item.concrete_path,
                            reasons=f_item.get_failure(),
                        )
                        failures.append(failure_item)

            self.filter = filtered_data
            self._tested = True

        else:
            self._is_valid = True
            self._tested = False
            self.filter = None

        self._failures = tuple(failures)



class RuleValidationFailure:
    def __init__(
        self,
        rule,
    ):
        pass


class DuplicateRule(Exception):
    pass


class IncompatibleRules(Exception):
    pass


class ValidationError(Exception):
    pass


class InvalidCallable(Exception):
    pass


class MalformedConditionLikeSpec(Exception):
    pass


class MalformedContainerItemSpec(Exception):
    pass


class MalformedDataPathSpec(Exception):
    pass


class MalformedRuleSpec(Exception):
    pass



import copy
from pathlib import Path

from ruamel.yaml import YAML
from valida.data import Data

from valida.rules import Rule


class Schema:
    def __init__(self, rules):
        self.rules = rules
        self.rule_tests = None  # Assigned by `self.validate`

    def __len__(self):
        return len(self.rules)

    def __eq__(self, other):
        if (
            type(self) == type(other)
            and self.rules == other.rules
            and self.rule_tests == other.rule_tests
        ):
            return True
        return False

    @staticmethod
    def init_rules(rules_dat):
        rules = []
        for i in rules_dat:
            rule = Rule.from_spec(i)
            rules.append(rule)
        return rules

    @classmethod
    def from_yaml(cls, yaml_str):
        yaml = YAML(typ="safe")
        schema_dat = yaml.load(yaml_str)
        rules = cls.init_rules(schema_dat["rules"])
        return cls(rules)

    @classmethod
    def from_yaml_file(cls, path):
        yaml = YAML(
            typ="safe"
        )  # need Python-native containers e.g. for rules that check container types
        with Path(path).open("r") as fh:
            schema_dat = yaml.load(fh)
        rules = cls.init_rules(schema_dat["rules"])
        return cls(rules)

    def validate(self, data):
        if not isinstance(data, Data):
            data = Data(data)
        return ValidatedData(self, data)


class ValidatedData:
    def __init__(self, schema, data):

        self.data = data
        self.schema = schema

        data_copy = copy.deepcopy(self.data.get_original())
        self.rule_tests = tuple(
            rule.test(self.data, _data_copy=data_copy) for rule in self.schema.rules
        )

        self.cast_data = data_copy

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"is_valid={self.is_valid!r}, "
            f"num_failures={self.num_failures}, "
            f"frac_rules_tested={self.frac_rules_tested}"
            f")"
        )

    @property
    def is_valid(self):
        return all(i.is_valid for i in self.rule_tests)

    @property
    def num_rules_tested(self):
        return sum(i.tested for i in self.rule_tests)

    @property
    def frac_rules_tested(self):
        return self.num_rules_tested / len(self.schema)

    @property
    def num_failures(self):
        return sum(i.num_failures for i in self.rule_tests)

    def get_failures_string(self):

        out = ""
        rules_tested_msg = (
            f"{self.num_rules_tested}/{len(self.schema)} rules were tested."
        )
        if self.is_valid:
            out += f"Data is valid. {rules_tested_msg}\n"
            return

        out += (
            f"{self.num_failures} rule{'s' if self.num_failures > 1 else ''} "
            f"failed validation. {rules_tested_msg}\n\n"
        )

        for rule_idx, rule_test in enumerate(self.rule_tests, start=1):
            if not rule_test.is_valid:
                rule_msg = f"Rule #{rule_idx}"
                out += rule_msg + "\n" + "-" * len(rule_msg) + "\n"
                out += rule_test.get_failures_string()
                out += "\n"

        return out

    def print_failures(self):
        print(self.get_failures_string())


class _TestDataSchema:
    def __init__(self, data, schema):

        rules = Schema.init_rules(schema["rules"])

        self.data = data
        self.schema = Schema(rules)

    @classmethod
    def from_yaml_file(cls, path):
        yaml = YAML(
            typ="safe"
        )  # need Python-native containers e.g. for rules that check container types
        with Path(path).open("r") as fh:
            all_dat = yaml.load(fh)
        test_data_schema = cls(all_dat["data"], all_dat["schema"])
        return test_data_schema

    @property
    def data_and_schema(self):
        return self.data, self.schema
